local M = {}

local tGroup = {} -- traffic group
-- both JBeam and config names (as long as they're in the right format) will work within these groups
tGroup.normal = {"autobello", "barstow", "burnside", "coupe", "etk800", "etkc", "etki", "fullsize", "hatch", "hopper", "legran", "midsize", "miramar", "moonhawk", "pessima", "pickup", "pigeon", "roamer", "sbr", "sunburst", "super", "van"}
tGroup.allBase = {}
tGroup.allAny = {}
tGroup.light = {"autobello/130_m.pc", "coupe", "hatch/DXi_M.pc", "midsize/LX_M.pc", "miramar", "pessima/HX_M.pc", "pigeon", "pigeon/boxed.pc"}
tGroup.heavy = {"citybus", "pickup/v8_4wd.pc", "pickup/beast.pc", "roamer/v8_4wd.pc", "roamer/offroad.pc", "semi", "semi/t75_dumptruck.pc", "van/medium.pc", "van/deliverytruck.pc"}
tGroup.america = {"barstow", "burnside", "coupe", "fullsize", "hatch", "legran", "midsize", "moonhawk", "pickup/ext_4wd.pc", "roamer", "sunburst", "van", "van/deliverytruck.pc"}
tGroup.europe = {"autobello", "autobello/130b_m.pc", "coupe", "etk800", "etkc", "etki", "etki/3000i_M_alt.pc", "hatch", "midsize", "pigeon", "super", "sbr"}
tGroup.asia = {"coupe", "hatch", "midsize", "miramar", "pessima", "pigeon", "sbr", "sunburst", "sunburst/custom.pc", "van/flatbed.pc"}
tGroup.americaPolice = {"coupe/police.pc", "fullsize/police.pc", "fullsize/bcpd.pc", "burnside/police.pc", "sunburst/police.pc", "sunburst/police_RS.pc", "roamer/sheriff.pc"}
tGroup.europePolice = {"super/polizia.pc", "etk800/police.pc", "etkc/police_highway_A.pc", "autobello/carabinieri.pc", "sunburst/gendarmerie.pc"}
tGroup.allPolice = {"coupe/police.pc", "super/polizia.pc", "etk800/police.pc", "fullsize/police.pc", "fullsize/bcpd.pc", "fullsize/unmarked.pc", "etkc/police_highway_A.pc", "autobello/carabinieri.pc", "burnside/police.pc", "sunburst/gendarmerie.pc", "sunburst/police.pc", "sunburst/police_RS.pc", "van/police.pc", "pickup/borderpatrol.pc", "roamer/sheriff.pc", "roamer/sheriff_offroad.pc"}
tGroup.automation = {}
tGroup.copyPlayer = {}
tGroup.custom = {}
tGroup.stock = {"autobello", "barstow", "bluebuck", "burnside", "citybus", "coupe", "etk800", "etkc", "etki", "fullsize", "hatch", "hopper", "legran", "midsize", "miramar", "moonhawk", "pessima", "pickup", "pigeon", "roamer", "sbr", "semi", "sunburst", "super", "van"}
tGroup.placeholder = {"pigeon/600.pc"}

local trafficAggro = 0.6 -- traffic aggression value
local racerAggro = 1.2 -- flee & chase mode aggression value
local policeSpawn = 1.2 -- police spawn multiplier at start of pursuit
local crawlSpeed = 3 -- slow vehicle speed
local arrestTime = 5 -- number of seconds until police arrest you
local evadeTime = 45 -- number of seconds away from police for evasion
local roadDrive = 0.1 -- minimum road drivability needed for spawn point
local nearDist = 40 -- near distance, used for traffic passing and interactions
local farDist = 400 -- far distance, used for traffic despawning
local pointDist = 200 -- start distance for point ahead of player vehicle; used for setting a spawn point
local readyDist = 100 -- proximity distance, used for a variety of traffic and police logic
local distMultiplier = 50 -- number used as a multiplier for despawn logic

local player = {}
local traffic = {}

zSiren = 0
local queue = 0
local loopTimer = 0
local timeTick = 0.1 + 1e-6
local speedLimit = 14
local speedConvert = {2.2369, 3.6} -- km/h, mph
local speedUnits = 2
local trafficTotal = 1
local spawnFactor = 1
local spawnGapFactor = 1
local distFactor = 1
local damageUnit = 1000
local up = vec3(0, 0, 1)
local group = tGroup.normal
local groupName = "normal"
local trafficMode = "random"
local laneMode = "on"
local paintColorMode = "standard"
local mainMap = nil
local metricSystem = true
local smartSpeed = true
local trafficReady = false
local policeReady = false
local policeAlert = false
local scatterCenter = true
local headlightsMode = true
local damageMode = true
local gearboxMode = core_settings_settings.getValue("defaultGearboxBehavior")

local function ready()
	log("A", "gull", "Traffic Tool ready")
end

local function getRoadScore(drivability, radius) -- gets the average value of road drivability and road radius
	return (drivability + (math.min(6 / 5, radius / 5))) / 2 -- limit for super wide roads
end

local function pickSpawn(spawnPointMid, spawnGap) -- picks three possible spawn points and gives them scores; best one wins (future: try path based spawn point)
	local spawnPoint = {pos = nil, rot = nil, data = nil, p1 = nil, p2 = nil, dist = nil}
	local bestPoint = {}
	local angles = {quatFromDir(vec3(-0.577, 1, 0):normalized(), up), quatFromDir(vec3(0, 1, 0), up), quatFromDir(vec3(0.577, 1, 0):normalized(), up)} -- 30 degrees
	local roadScore = 0
	local score = 0
	local bestScore = 0
	local upScore = 0
	local a = 0
	if scatterCenter == true then -- randomizes or shifts the center point
		if mainMap.nodes[spawnPointMid.p1].links[spawnPointMid.p2].oneWay == true then
			a = 0.08 -- right side bias; makes highways a bit more fun
		else
			a = math.random(-132, 132) / 1000 -- random narrow angle
		end
		angles[2] = quatFromDir(vec3(a, 1, 0):normalized(), up)
	end
	for i, v in ipairs(angles) do
		spawnPoint.p1, spawnPoint.p2, spawnPoint.dist = map.findClosestRoad(vec3(spawnPointMid.pos) - vec3(spawnPointMid.rot):rotated(angles[i]) * spawnGap * -1)
		if i == 2 then -- center spawn point is favored and therefore given a small multiplier
			upScore = 1.15
		else
			upScore = 1
		end
		if spawnPoint.p1 then
			spawnPoint.data = mainMap.nodes[spawnPoint.p1].links[spawnPoint.p2]
			roadScore = getRoadScore(spawnPoint.data.drivability, mainMap.nodes[spawnPoint.p1].radius)
			score = math.ceil((100 - (spawnPoint.dist / 10)) * roadScore * upScore) -- formula for spawn point score
		end
		if score >= bestScore then
			bestScore = score
			for k, v in pairs(spawnPoint) do
				bestPoint[k] = v
			end
		end
	end
	return bestPoint
end

local function findSpawn(t) -- finds and returns a spawn point on the map
	local spawnPoint = {}
	local spawnPointMid = {}
	local gotSpawn = false
	local spawnGap = nil
	local loopCount = 1
	while gotSpawn == false do
		if loopCount == 1 then
			spawnPointMid.p1, spawnPointMid.p2, spawnPointMid.dist = map.findClosestRoad(vec3(player.pos) - vec3(player.vec) * (pointDist / 2) * -1) --Midway spawn point focus
		elseif loopCount >= 30 then -- 30 loops
			spawnPoint.p1, spawnPoint.p2, spawnPoint.dist = nil, nil, nil
			break
		end
		if spawnPointMid.p1 then
			spawnPointMid.pos = mainMap.nodes[spawnPointMid.p1].pos
			spawnPointMid.rot = (vec3(spawnPointMid.pos) - vec3(player.pos)):normalized()
			spawnGap = (pointDist / 2) * spawnGapFactor * square(distFactor) + (player.vel * 2)  + ((loopCount - 1) * 25) -- higher player velocity moves the spawn point further; loop moves it even further
			spawnPoint = pickSpawn(spawnPointMid, spawnGap)
		end
		if spawnPoint.p1 then
			gotSpawn = true -- temporarily set to true
			spawnPoint.pos = mainMap.nodes[spawnPoint.p1].pos
			spawnPoint.data = mainMap.nodes[spawnPoint.p1].links[spawnPoint.p2]
			for i, v in ipairs(traffic) do
				if traffic[i].obj.name ~= t.obj.name then
					if (spawnPoint.pos:toPoint3F() - traffic[i].pos):len() <= pointDist / 8 or spawnPoint.data.drivability < roadDrive then -- prevents spawn points inside other traffic cars or non roads
						gotSpawn = false
					end
				end
			end
			if (spawnPoint.pos:toPoint3F() - player.pos):len() <= pointDist * spawnGapFactor then -- prevents spawn points too close to player
				gotSpawn = false
			end
		end
		
		loopCount = loopCount + 1
	end
	return spawnPoint
end

local function findParallelRoad(spawnPoint, spawnGap) -- finds parallel roads to the left or right of the original road
	local tempPoint = {}
	local point1 = {}
	local point2 = {}
	local roadDir = 0
	if not spawnGap then
		spawnGap = 5
	end
	local spawnPosTemp = {mainMap.nodes[spawnPoint.p1].pos, mainMap.nodes[spawnPoint.p2].pos}
	local spawnGapTemp = {(player.pos - spawnPosTemp[1]:toPoint3F()):len(), (player.pos - spawnPosTemp[2]:toPoint3F()):len()}
	if spawnGapTemp[1] < spawnGapTemp[2] then -- figures out which spawn point from the pair is closest to the player so a proper rotation can be done
		tempPoint.pos = spawnPosTemp[1]
		tempPoint.rot = (vec3(spawnPosTemp[2]) - vec3(tempPoint.pos)):normalized()
	else
		tempPoint.pos = spawnPosTemp[2]
		tempPoint.rot = (vec3(tempPoint.pos) - vec3(spawnPosTemp[1])):normalized()
	end
	for i = 1, 2, 1 do
		if i == 1 then
			point1.pos = vec3(tempPoint.pos) - tempPoint.rot:z0():cross(up) * spawnGap * -1
		else
			point1.pos = vec3(tempPoint.pos) - tempPoint.rot:z0():cross(up) * spawnGap
		end
		point1.p1, point1.p2, point1.dist = map.findClosestRoad(point1.pos)
		point1.rot = (vec3(mainMap.nodes[point1.p1].pos) - vec3(mainMap.nodes[point1.p2].pos)):normalized()
		if point1.dist <= spawnGap / 1.3 and math.abs(tempPoint.rot:dot(point1.rot)) >= 0.8 then -- nearby side node found with similar direction
			roadDir = i
		else
			point2.pos = vec3(point1.pos) - tempPoint.rot * spawnGap * 1.5 * -1 -- slight distance boost
			point2.p1, point2.p2, point2.dist = map.findClosestRoad(point2.pos)
			point2.rot = (vec3(mainMap.nodes[point2.p1].pos) - vec3(mainMap.nodes[point2.p2].pos)):normalized()
			if point2.dist <= spawnGap / 1.3 and point1.p1 ~= point2.p1 and math.abs(point1.rot:dot(point2.rot)) >= 0.8 then -- nearby diagonal node found with similar direction
				roadDir = i
			end
		end
	end
	return roadDir, quatFromDir(tempPoint.rot, up)
end

local function setSpawn(t) -- sets the new spawn point of a traffic car and rotates it logically
	local spawn = {dir = 1, highway = 0, shiftOut = 0, shiftUp = 0.5}
	local spawnPoint = {pos = nil, rot = nil, data = nil, p1 = nil, p2 = nil, dist = nil}
	local laneShift = 1.75
	local colorSet = {}
	local colorValues = {}
	spawnPoint = findSpawn(t)
	if not spawnPoint.p1 then
		log("A", "gull", "Warning: "..t.obj.name.." ("..t.obj.jbeam..") didn't get a new spawn point!")
	else
		spawn.pos = spawnPoint.pos
		spawn.n1 = mainMap.nodes[spawnPoint.p1]
		spawn.n2 = mainMap.nodes[spawnPoint.p2]
		spawn.roadScore = getRoadScore(spawnPoint.data.drivability, spawn.n1.radius)
		spawn.highway, spawn.rot = findParallelRoad(spawnPoint, 24)
		t.lane = 1
		if spawnPoint.data.oneWay == false then
			spawn.dir = math.random(1, 2)
			if spawn.highway > 0 then
				spawn.dir = spawn.highway
			else
				spawn.rot = quatFromDir(spawn.n1.pos - spawn.n2.pos, up) -- best way to get node rotation
			end
			if spawn.n1.radius <= 2.2 then -- road is really narrow (temporary solution)
				spawn.rot = quatFromDir(player.vec)
				spawn.dir = 2
			end
		else
			if spawnPoint.data.inNode == spawnPoint.p1 then
				spawn.rot = quatFromDir(spawn.n1.pos - spawn.n2.pos, up)
			else
				spawn.rot = quatFromDir(spawn.n2.pos - spawn.n1.pos, up)
			end
			if spawn.n1.radius >= 3 then -- road is wide enough to have at least two lanes
				t.lane = math.random(0, 1)
			end
		end
		if t.lane == 1 then
			spawn.shiftOut = laneShift * spawn.roadScore * (spawn.dir * -2 + 3)
		else
			spawn.shiftOut = laneShift * spawn.roadScore * (t.lane * 2 - 1)
		end
		spawn.posA = vec3(spawn.pos:toPoint3F()) - vec3(0, 1, 0):rotated(spawn.rot):z0():cross(up) * spawn.shiftOut
		if spawn.dir == 1 then
			spawn.rotA = quatFromDir(vec3(0, 1, 0):rotated(spawn.rot), up)
		else
			spawn.rotA = quatFromDir(vec3(0, -1, 0):rotated(spawn.rot), up)
		end
		t.obj:setPositionRotation(spawn.posA.x, spawn.posA.y, spawn.posA.z + spawn.shiftUp, spawn.rotA.x, spawn.rotA.y, spawn.rotA.z, spawn.rotA.w)
		log("A", "gull", t.obj.name.." ("..t.obj.jbeam..") teleported to "..spawn.posA.x..", "..spawn.posA.y..", "..spawn.posA.z)
		t.obj:queueLuaCommand("vehicleResetted(false)")
		t.obj:resetBrokenFlexMesh()
		if paintColorMode ~= "unchanged" and t.role ~= "police" and t.role ~= "ambulance" then
		local color = {}
			if paintColorMode == "standard" then
				color = t.colors[math.random(1, #t.colors)]
			elseif paintColorMode == "random" then
				color = {math.random(), math.random(), math.random(), math.random() + 1}
			end
			t.obj.color = ColorF(color[1], color[2], color[3], color[4]):asLinear4F()
			if t.obj.jbeam == "fullsize" or t.obj.jbeam == "roamer" then -- stock vehicles with dual paint styles
				t.obj.colorPalette0 = t.obj.color
			end
		end
		t.road.drive = spawnPoint.data.drivability
		t.road.oneWay = spawnPoint.data.oneWay
		t.road.radius = mainMap.nodes[spawnPoint.p1].radius
		t.road.score = spawn.roadScore
		t.road.parallel = spawnHighway
	end
	t.state = "set"
	t.delay = false
end

local function getNumberList(a, b) -- returns a list of non-repeating random integers
	local n = {}
	local l = {}
	local key = 0
	for i = 1, b, 1 do
		table.insert(n, i)
	end
	for i = 1, a, 1 do
		key = math.random(1, #n)
		table.insert(l, n[key])
		table.remove(n, key)
	end
	return l
end

local function setQuantity(value)
	if not value then
		log("A", "gull", "Quantity was nil!")
	else
		trafficTotal = value
	end
end

local function setGroup(int)
	local groupList = {"normal", "allBase", "allAny", "light", "heavy", "america", "europe", "asia", "americaPolice", "europePolice", "allPolice", "automation", "copyPlayer", "custom"}
	groupName = groupList[int]
	for k, v in pairs(tGroup) do
		if groupName == k then
			group = v
			ui_message("Spawn group set: "..groupName, 6)
			log("A", "gull", "Group: "..groupName)
		end
	end
	if int > tableSize(groupList) then
		log("A", "gull", "Group ID doesn't exist!")
	end
end

local function setCustom(list) -- sets a custom traffic group for the spawner
	local tList = {}
	if type(list) == "string" then
		tList = {list}
	elseif type(list) == "table" then
		tList = list
	else
		tList = tGroup.placeholder
	end
	tGroup.custom = tList
	group = tGroup.custom
	ui_message("Custom data ready", 6)
	log("A", "gull", "Custom data ready")
end

local function setAsPolice(config) -- sets the current player vehicle to act as police
	if not config then
		config = "nil"
		local tempPlayer = be:getPlayerVehicle(0)
		if tempPlayer.partConfig then
			config = string.gsub(tempPlayer.partConfig, "vehicles/", "")
		end
	end
	table.insert(tGroup.allPolice, config)
	ui_message("Police car set: "..config, 6)
	log("A", "gull", "Police vehicle: "..config)
end

local function togglePolice()
	policeReady = not policeReady
	log("A", "gull", "Police mode: "..tostring(policeReady))
end

local function toggleHeadlights()
	headlightsMode = not headlightsMode
	log("A", "gull", "Headlights mode: "..tostring(headlightsMode))
end

local function toggleDamageKill()
	damageMode = not damageMode
	log("A", "gull", "Damage mode: "..tostring(damageMode))
end

local function setSpeedLimit(value)
	if not value then
		log("A", "gull", "Speed limit value was nil!")
	else
		speedLimit = value / speedConvert[speedUnits] -- to m/s
	end
end

local function toggleUnits(value)
	if not value then
		log("A", "gull", "Unit value was nil!")
	else
		speedUnits = value + 1
		setSpeedLimit(speedLimit * speedConvert[3 - speedUnits]) -- sloppy
	end
end

local function setRespawnRate(value)
	if not value then
		log("A", "gull", "Respawn rate value was nil!")
	else
		spawnFactor = value
		spawnGapFactor = -value / 5 + 1.2 -- variable respawn distance multiplier
	end
end

local function setColorMode(value)
	if not value then
		log("A", "gull", "Color mode value was nil!")
	else
		if value == 1 then
			paintColorMode = "unchanged"
		elseif value == 2 then
			paintColorMode = "standard"
		elseif value == 3 then
			paintColorMode = "random"
		end
	end
end

local function setAiMode(value)
	if not value then
		log("A", "gull", "AI mode value was nil!")
	else
		if value == 1 then
			trafficMode = "random"
			laneMode = "on"
		elseif value == 2 then
			trafficMode = "flee"
			laneMode = "off"
		elseif value == 3 then
			trafficMode = "chase"
			laneMode = "off"
		elseif value == 4 then
			trafficMode = "disabled"
			laneMode = "off"
		end
		if trafficReady == true then
			for i, v in ipairs(traffic) do
				traffic[i].obj:queueLuaCommand("aitt.setMode('"..trafficMode.."')")
				traffic[i].obj:queueLuaCommand("aitt.driveInLane('"..laneMode.."')")
				if trafficMode == "disabled" then
					traffic[i].obj:queueLuaCommand("extensions.unload('aitt')")
					trafficReady = false
				end
			end
		end
	end
end

local function toggleAwareness(value)
	if not value then
		log("A", "gull", "Awareness mode value was nil!")
	else
		if value == 1 then
			smartSpeed = true
		else
			smartSpeed = false
		end
	end
end

local function toggleMethod(value)
	if not value then
		log("A", "gull", "Spawn method mode value was nil!")
	else
		if value == 1 then
			distFactor = 1.4
		else
			distFactor = 1
		end
	end
end

local function getAllVehicles(class) -- gets all vehicles depending on group type
	if not class then
		class = "allAny"
	end
    local vList = {}
	local name = nil
	local json = nil
    for i, v in ipairs(FS:findFiles("/vehicles", "*", 0, false, true)) do
        if v ~= "/vehicles/common" then
			json = jsonReadFile(v.."/info.json")
			if json then
				if json["Type"] == "Car" or json["Type"] == "Truck" or json["Type"] == "Automation" then
					if class == "allAny" or class == "custom" then
						for o, w in ipairs(FS:findFiles(v, "*.pc", 0, true, false)) do
							if class == "allAny" then
								table.insert(vList, w)
							else
								local stock = false
								for u, x in ipairs(tGroup.stock) do -- no match in stock vehicles
									if string.find(string.lower(w), x) then
										stock = true
									end
								end
								if stock == false then
									table.insert(vList, w)
								end
							end
						end
					else	
						if string.byte(v) == 47 then
							name = string.gsub(v, "/vehicles/", "")
						else
							name = string.gsub(v, "vehicles/", "")
						end
						name = string.match(name, "%w*_*%w*_*%w*_*%w*_*%w*") -- this seems bad
						if class == "allBase" then
							table.insert(vList, name)
						elseif class == "automation" then
							if json["Type"] == "Automation" then
								table.insert(vList, name)
							end
						end
					end
				end
			end
        end
    end
	return vList
end

local function checkVehicle(vehicle) -- gets the vehicle type from the .json file
	local vType = nil
	local jbName = scenetree.findObject(vehicle).jbeam
	local json = jsonReadFile("/vehicles/"..jbName.."/info.json")
	if json then
		if json["Type"] then
			vType = string.lower(json["Type"])
		end
	else
		vType = "unknown" -- hmm
	end
	return vType
end

local function getColors(jbeam) -- returns a vehicle's paint color choices
	local json = jsonReadFile("/vehicles/"..jbeam.."/info.json")
	local colors, colorTable = {}, {}
	if json then
		colors = json["colors"]
	else
		colors = {white = "1 1 1 1.5"}
	end
	for k, v in pairs(colors) do
		table.insert(colorTable, stringToTable(colors[k], " ")) -- {1, 1, 1, 1.5}
	end
	return colorTable
end

local function getTrafficTable() -- returns a table with valid traffic vehicles
	local otherVehicles = scenetree.findClassObjects("BeamNGVehicle")
	local validTypes = {"car", "truck", "automation"}
	local delete = {}
	for i, v in ipairs(otherVehicles) do
		local vType = checkVehicle(v)
		local isValid = false
		for o, w in ipairs(validTypes) do
			if vType == w then
				isValid = true
			end
		end
		if otherVehicles[i] == be:getPlayerVehicle(0).name or isValid == false then
			table.insert(delete, i)
		end
	end
	for i = #delete, 1, -1 do
		table.remove(otherVehicles, delete[i])
	end
	return otherVehicles
end

local function findClosestTraffic() -- finds the nearest traffic vehicle
	local tempDist = farDist
	local t = nil
	for i, v in ipairs(traffic) do
		if traffic[i].dist.now <= tempDist then
			tempDist = traffic[i].dist.now
			t = traffic[i]
		end
	end
	return t
end

local function validateVehicles() -- validates all vehicles; prevents game crashes and stuff
	if trafficReady == true then
		local otherVehicles = getTrafficTable()
		if #otherVehicles < #traffic and #traffic > 0 then
			traffic = {}
			for i, v in ipairs(otherVehicles) do
				local t = {}
				t.obj = scenetree.findObject(v)
				t.colors = getColors(t.obj.jbeam)
				t.state = "set"
				t.role = "normal"
				t.action = "none"
				t.speed = speedLimit
				t.lane = 1
				t.dist = {}
				t.damage = {}
				t.road = {}
				t.node = {}
				t.timer = {wait = 0, active = 0, path = 0, delay = false}
				t.obj:queueLuaCommand("extensions.load('aitt')")
				table.insert(traffic, t)
			end
		else
			if #otherVehicles == 0 then
				trafficReady = false
			end
		end
	end
end

local function deleteVehicles() -- deletes traffic vehicles
	local otherVehicles = getTrafficTable()
	local quantity = math.min(#otherVehicles, trafficTotal)
	for i = #otherVehicles, #otherVehicles - quantity + 1, -1 do
		scenetree.findObject(otherVehicles[i]):delete()
	end
	validateVehicles()
	ui_message(quantity.." vehicles deleted", 6)
end

local function setInitSpawn(active, total) -- sets the spawn points of the initial vehicles
	local list = {}
	if total <= #group then
		list = getNumberList(total, #group)
	else
		for i = 1, total, 1 do
			table.insert(list, math.random(1, #group))
		end
	end
	local groupData = {}
	local inData = {}
	local player0 = be:getPlayerVehicle(0)
	local mainColor = nil
	local spawnGap = 12
	if groupName == "heavy" then
		spawnGap = 16
	end
	if groupName == "copyPlayer" then
		tGroup.copyPlayer = {player0.partConfig}
		mainColor = tostring(player0.color)
		mainColor = stringToTable(mainColor, ", ")
		mainColor = ColorF(mainColor[1], mainColor[2], mainColor[3], mainColor[4]) -- phew! Is there a better way to do this?
		for i = 1, total, 1 do
			table.insert(groupData, {player0.jbeam, player0.partConfig, mainColor})
		end
	else
		for i = 1, total, 1 do
			inData = {}
			local name = group[list[i]]
			local jbName = nil
			local nameSpec = nil
			local json = nil
			mainColor = "1 1 1 1.5"
			if string.byte(name) == 47 then -- temporarily remove "/vehicles/" from string
				name = string.gsub(name, "/vehicles/", "")
			else
				name = string.gsub(name, "vehicles/", "")
			end
			if string.find(name, ".pc") then
				nameSpec = "pc"
			end
			if nameSpec == "pc" then
				jbName = string.match(name, "%w*_*%w*_*%w*_*%w*_*%w*")
				inData[1] = jbName
				inData[2] = "vehicles/"..name
			else
				jbName = name
				inData[1] = jbName
				inData[2] = nil
			end
			json = jsonReadFile("/vehicles/"..jbName.."/info.json")
			if json then
				local colorTable = json["colors"]
				local colorName = json["default_color"]
				local fileName = nil
				if nameSpec == "pc" then
					name = string.gsub(name, jbName.."/", "")
					name = string.gsub(name, ".pc", "")
					fileName = "/vehicles/"..jbName.."/info_"..name..".json"
					json = jsonReadFile(fileName)
					if json then
						colorName = json["default_color"]
					end
				end
				for k, v in pairs(colorTable) do
					if colorName == k then
						mainColor = v
					end
				end
			end
			mainColor = stringToTable(mainColor, " ")
			inData[3] = ColorF(mainColor[1], mainColor[2], mainColor[3], mainColor[4])
			table.insert(groupData, inData)
		end
	end
	for i = 1, total, 1 do
		local spawnPos = vec3(player0:getPosition()) - vec3(player0:getDirectionVector()) * spawnGap * (active + i) * -1
		local spawnRot = vec3(player0:getDirectionVector())
		spawn.spawnVehicle(groupData[i][1], groupData[i][2], vec3(spawnPos):toPoint3F(), quatFromDir(spawnRot, up), groupData[i][3])
	end
end

local function spawnVehicles() -- spawns the required amount of vehicles
	if trafficTotal >= 1 then
		local player0 = be:getPlayerVehicle(0)
		local otherVehicles = getTrafficTable()
		if groupName == "allBase" or groupName == "allAny" or groupName == "automation" or groupName == "custom" then
			tGroup[groupName] = getAllVehicles(groupName)
			group = tGroup[groupName]
			if #group == 0 then
				group = tGroup.placeholder
			end
		end
		local trafficNeeded = trafficTotal - #otherVehicles
		if trafficNeeded > 0 then
			setInitSpawn(#otherVehicles, trafficNeeded)
			be:enterVehicle(0, player0)
			if trafficNeeded >= 12 then
				ui_message("Spawned "..trafficNeeded.." new AI vehicles; that's quite insane!", 12)
			else
				ui_message("Spawned "..trafficNeeded.." new AI vehicles", 6)
			end
		else
			ui_message("There are already "..trafficTotal.." AI vehicles on the map!", 6)
		end
	else
		log("A", "gull", "Traffic quantity set to 0!")
	end
end

local function activateVehicles() -- processes and activates vehicles on the map, and enables the main script to run
	mainMap = map.getMap()
	player = {}
	traffic = {}
	player.obj = be:getPlayerVehicle(0)
	player.pair = {name = nil, pair = false}
	player.timer = {police = 0, evade = 0}
	local otherVehicles = getTrafficTable()
	if #otherVehicles < 1 then
		trafficReady = false
	else
		for i, v in ipairs(otherVehicles) do
			local t = {}
			t.obj = scenetree.findObject(v)
			t.colors = getColors(t.obj.jbeam)
			t.state = "set"
			t.role = "normal"
			t.action = "none"
			t.speed = speedLimit
			t.lane = 1
			t.dist = {}
			t.damage = {}
			t.road = {}
			t.node = {}
			t.timer = {wait = 0, active = 0, path = 0, delay = false}
			t.obj:queueLuaCommand("extensions.load('aitt')")
			table.insert(traffic, t)
		end
		trafficReady = true
		policeReady = false
		for i, t in ipairs(traffic) do
			t.obj:queueLuaCommand("aitt.setSpeedMode('limit')")
			t.obj:queueLuaCommand("aitt.resetLearning(true)")
			for o, w in ipairs(tGroup.allPolice) do
				if t.obj.partConfig == "vehicles/"..tGroup.allPolice[o] then
					t.role = "police"
					policeReady = true
				end
			end
			if string.find(traffic[i].obj.partConfig, "police") then
				t.role = "police"
				policeReady = true
			end
			if string.find(traffic[i].obj.partConfig, "ambulance") then
				t.role = "ambulance"
			end
		end
		ui_message("Traffic activated!", 6)
	end
end

local function getSmartSpeed(t) -- sets a situational speed or action for the AI (needs more work)
	local speed = 0
	local mode = "normal"
	local target = nil
	local policeDist = farDist
	local randomAction = 0
	local step = 10
	local speedWatch = 10
	local speedFactor = 0
	local tempDist = 0
	local pointDist = (2.5 + math.min(1.5, t.vel / 5)) * t.road.score -- point area changes
	if player.pair.lock == true and t.lane == 1 then -- experimental thing here
		pointDist = 2 * t.road.score
		t.lane = 0.8
		t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
	end
	local frontDist = 5
	local frontValue = math.max(1, t.vel)
	local frontPos = nil
	local tempPos = vec3(t.pos)
	for i = 2, 0, -1 do
		frontPos = (tempPos - t.vec * (frontDist + frontValue * i) * -1):toPoint3F() -- AI front position points are done in "beads"
		for o, w in ipairs(traffic) do
			if traffic[o].obj.name ~= t.obj.name then
				tempDist = (frontPos - traffic[o].pos):len()
				if tempDist <= pointDist then
					step = i
					target = "traffic"
					speedWatch = traffic[o].vel
				end
			end
		end
		tempDist = (frontPos - player.pos):len()
		if tempDist <= pointDist and i < step then
			step = i
			target = "player"
			speedWatch = player.vel
		end
	end
	if step <= 2 then
		if target == "player" and player.vec:dot(t.vec) <= -0.9 then -- player facing AI head to head; AI will brake a bit harder
			speedWatch = 0
		end
		speedFactor = 0.75 + (step / 15) + (math.max(0.1, speedWatch) / math.max(0.1, t.vel)) / 6 -- sloppy formula here
		speed = math.min(speedLimit, (speedWatch * speedFactor)) -- formula for distance based speed
	else
		if t.timer.wait > 0 then
			if t.speed < 1 then
				speed = t.speed + math.min(0.15, speedWatch / 10) -- somewhat gradual acceleration
			else
				speed = t.speed * 1.08
			end
		else
			if t.lane == 0 then
				speed = speedLimit * t.road.score * 1.15
			else
				speed = speedLimit * t.road.score
			end
		end
	end
	for i, v in ipairs(traffic) do
		if traffic[i].obj.name ~= t.obj.name then
			if t.road.radius <= 2.5 and t.road.parallel == 0 then
				if ((t.node[2]:toPoint3F() - traffic[i].node[2]:toPoint3F()):len() <= 5 or (t.node[2]:toPoint3F() - player.pos):len() <= 10) then
					mode = "pullOver"
				end
			end
			if t.role ~= "police" and traffic[i].role == "police" and traffic[i].action ~= "none" then
				if (t.node[3]:toPoint3F() - traffic[i].pos):len() <= policeDist then
					policeDist = (t.node[3]:toPoint3F() - traffic[i].pos):len()
				end
			end
		end
	end
	if zSiren > 0 then -- if player siren is on
		local frontGap = readyDist / 2
		local tempObj = nil
		frontPos = vec3(player.pos) - player.vec * frontGap * -1
		tempDist = (t.pos - frontPos:toPoint3F()):len()
		if player.vel > crawlSpeed then
			player.pair.name = t.obj.name
			player.pair.lock = false
		else
			if player.pair.lock == false then
				tempObj = findClosestTraffic()
				player.pair.name = tempObj.obj.name
				player.pair.lock = true
			end
		end
		if t.role ~= "police" and player.pair.name == t.obj.name then
			if tempDist <= frontGap then
				policeDist = tempDist
				if t.timer.path == 0 and math.random(1, 10) == 1 then
					t.role = "suspect"
				end
			end
		end
	end
	if policeDist <= 50 then
		mode = "pullOverPolice"
	else
		if t.timer.path ~= 0 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
	end
	if mode == "bypass" then -- modes use the path timer to process actions (better method needed)
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 0.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 5 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
	elseif mode == "pullOver" then
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 1.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 5 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
		speed = math.max(crawlSpeed, speed - t.timer.path * 1.5)
	elseif mode == "pullOverPolice" then
		if t.timer.path == 0 then
			t.timer.path = 0.1
			t.lane = 1.5
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
		elseif t.timer.path >= 8 then
			if t.role == "suspect" then -- traffic vehicle runs from police!
				t.action = "flee"
				t.obj:queueLuaCommand("aitt.setMode('flee')")
				t.obj:queueLuaCommand("aitt.driveInLane('off')")
				t.respawn = 0.1
				t.timer.path = 0
			end
		elseif t.timer.path >= 60 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.timer.path = 0
		end
		speed = math.max(0.01, speed - t.timer.path * 3)
	end
	if t.timer.path ~= 0 then
		t.timer.path = t.timer.path + timeTick
	end
	return speed
end

local function updateRoad(t) -- updates the road data of the vehicle
	local point = {}
	point.p1, point.p2, point.dist = map.findClosestRoad(vec3(t.node[1]))
	if point.p1 then
		t.road.drive = mainMap.nodes[point.p1].links[point.p2].drivability
		t.road.oneWay = mainMap.nodes[point.p1].links[point.p2].oneWay
		t.road.radius = mainMap.nodes[point.p1].radius
		t.road.score = getRoadScore(t.road.drive, t.road.radius)
		t.road.parallel = findParallelRoad(point, 12)
		if t.lane <= 0 then
			if not t.road.oneWay or (t.road.oneWay and t.road.radius < 3) then -- prevents odd one way road behaviors
				t.lane = 1
				t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			end
		end
	end
end

local function chasePlayer(t) -- new method for AI to chase player
	local frontGap = readyDist / 2
	local frontPos = vec3(t.pos) - t.vec * frontGap * -1 -- point ahead of vehicle; chase mode trickery
	local frontDist = (player.pos - frontPos:toPoint3F()):len()
	if t.action == "watch" then
		t.speed = math.max(crawlSpeed, player.vel * ((readyDist / math.min(readyDist, t.dist.now)) / 15 + 0.4)) -- formula for "chase follow" speed
		if frontDist <= frontGap + math.max(5, player.vel / 4) or player.vel <= crawlSpeed then
			if t.role == "police" then
				t.respawn = math.max(0, policeSpawn)
			else
				t.respawn = 1
			end
			t.speed = 100
			t.action = "chase"
			t.obj:queueLuaCommand("aitt.driveInLane('off')")
			t.obj:queueLuaCommand("aitt.setMode('chase')")
		end
	end
end

local function simulatePolice() -- various logic for while the police car is chasing the player
	local chaseDist = readyDist / 2
	local arrestDist = farDist
	for i, v in ipairs(traffic) do
		if traffic[i].role == "police" and traffic[i].action ~= "none" then
			if traffic[i].dist.now <= arrestDist then
				arrestDist = traffic[i].dist.now
			end
		end
	end
	if player.timer.police >= arrestTime then
		arrestDist = 0
	end
	if arrestDist <= chaseDist then 
		if player.vel <= crawlSpeed then
			if math.floor(player.timer.police * 10) / 10 == arrestTime then
				player.obj:queueLuaCommand("ai.setMode('stopping')")
				for i, v in ipairs(traffic) do
					if traffic[i].role == "police" then
						traffic[i].obj:queueLuaCommand("aitt.setMode('stopping')")
						traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(1)")
					end
				end
				ui_message("You got busted!", 5)
			elseif math.floor(player.timer.police * 10) / 10 == arrestTime + 5 then
				player.obj:queueLuaCommand("ai.setMode('disabled')")
				player.obj:queueLuaCommand("controller.mainController.setGearboxMode("..gearboxMode..")")
				for i, v in ipairs(traffic) do
					if traffic[i].role == "police" then
						traffic[i].obj:queueLuaCommand("aitt.setMode('random')")
						traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(0)")
						traffic[i].obj:queueLuaCommand("aitt.driveInLane('on')")
						traffic[i].speed = speedLimit
						traffic[i].action = "none"
					end
				end
				ui_message("You can drive away now...", 5)
				policeSpawn = 1.2
				policeAlert = false
			end	
			player.timer.police = player.timer.police + timeTick
		else
			player.timer.police = 0
		end
		player.timer.evade = 0
	else
		if player.timer.evade >= evadeTime then
			for i, v in ipairs(traffic) do
				if traffic[i].role == "police" then
					traffic[i].obj:queueLuaCommand("aitt.setMode('random')")
					traffic[i].obj:queueLuaCommand("electrics.set_lightbar_signal(0)")
					traffic[i].obj:queueLuaCommand("aitt.driveInLane('on')")
					traffic[i].speed = speedLimit
					traffic[i].action = "none"
				end
			end
			ui_message("You successfully evaded the police!", 5)
			policeSpawn = 1.2
			policeAlert = false
		end
	end
	player.timer.evade = player.timer.evade + timeTick
end

local function simulateTraffic() -- various logic for traffic; also handles when to respawn traffic
	player.pos = player.obj:getPosition()
	player.vel = map.objects[player.obj:getID()].vel:length()
	player.vec = vec3(player.obj:getDirectionVector())
	player.damage = map.objects[player.obj:getID()].damage
	player.obj:queueLuaCommand("obj:queueGameEngineLua('zSiren = ('..electrics.values.lightbar..')')") -- player siren state
	local timeDay = core_environment.getTimeOfDay()
	for i, v in ipairs(traffic) do
		local t = traffic[i]
		t.pos = t.obj:getPosition()
		t.dist.now = (player.pos - t.pos):len()
		t.vel = map.objects[t.obj:getID()].vel:length()
		t.vec = vec3(t.obj:getDirectionVector())
		t.damage.now = map.objects[t.obj:getID()].damage
		t.node[1] = vec3(t.pos) - vec3(t.vec) * 5 * -1 -- temporary thing
		t.node[2] = vec3(t.pos) - vec3(t.vec) * 20 * -1
		t.node[3] = vec3(t.pos) - vec3(t.vec) * 50 * -1
		if t.state == "set" then
			if t.timer.delay == true then -- slight one tick delay for AI mode change
				if trafficMode == "chase" then
					if t.role == "police" then
						policeAlert = true
					end
					t.obj:queueLuaCommand("aitt.setMode('random')")
					t.obj:queueLuaCommand("aitt.driveInLane('on')")
					t.action = "watch"
				else
					t.obj:queueLuaCommand("aitt.setMode('"..trafficMode.."')")
					t.obj:queueLuaCommand("aitt.driveInLane('"..laneMode.."')")
				end
				if trafficMode == "random" then
					t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
					t.obj:queueLuaCommand("aitt.setAggression("..tostring(trafficAggro)..")")
				else
					t.obj:queueLuaCommand("aitt.setAggression("..tostring(racerAggro)..")")
				end
				if t.role == "suspect" then
					t.role = "normal"
					t.action = "none"
				end
				if t.role == "police" and t.action ~= "none" then
					policeSpawn = policeSpawn - 0.02 -- police respawn multiplier decreases every time a new unit appears
					t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
					t.obj:queueLuaCommand("aitt.setAggression("..tostring(racerAggro)..")")
					t.speed = speedLimit
					t.lane = 0.8
					t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
					t.action = "watch"
				end
				if timeDay.time then
					if headlightsMode == true and timeDay.time >= 0.22 and timeDay.time <= 0.77 then
						t.obj:queueLuaCommand("electrics.setLightsState(1)")
					else
						t.obj:queueLuaCommand("electrics.setLightsState(0)")
					end
				end
				t.obj:queueLuaCommand("electrics.set_warn_signal(0)")
				t.state = "normal"
				t.timer = {wait = 0, active = 0, path = 0, delay = false}
			end
			updateRoad(t)
			t.respawn = 1
			t.damage.bad = false
			t.dist.prev = t.dist.now
			t.timer.delay = true
		end
		if t.timer.active < 1 then -- had to really force these variables
			t.damage.init = t.damage.now
			t.damage.prev = t.damage.now
		end
		if t.timer.active >= 1 and t.dist.now <= nearDist then
			if t.action == "none" then
				t.respawn = 1.5
			end
		end
		if t.dist.now <= readyDist / 2 then
			if t.role == "police" and t.action == "none" and policeReady == true then
				if (player.vel >= speedLimit * 1.25 and player.vel <= 100) or (t.dist.now <= nearDist / 2 and t.damage.now - t.damage.prev >= damageUnit / 2) then -- minor offense
					t.obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
					t.action = "watch"
					policeAlert = true
				end
				if (player.vel >= speedLimit * 1.4 and player.vel <= 100) or (t.dist.now <= nearDist / 2 and t.damage.now - t.damage.prev >= damageUnit * 2) then -- major offense
					for o, w in ipairs(traffic) do
						if traffic[o].role == "police" and traffic[o].action == "none" then
							traffic[o].obj:queueLuaCommand("electrics.set_lightbar_signal(2)")
							traffic[o].action = "watch"
							policeAlert = true
						end
					end
				end
			end
			t.damage.prev = t.damage.now
		end
		if spawnFactor == 0 then
			t.respawn = 0
		end
		local respawnValue = farDist * distFactor - ((t.dist.now - t.dist.prev) * (1 - math.max(0, (farDist / 2 - t.dist.now) / farDist)) * distMultiplier * spawnFactor * t.respawn) -- formula for smart traffic respawning
		if t.respawn > 0 and t.timer.active >= 1 and t.dist.now >= math.max(readyDist / 2, respawnValue) then
			t.state = "ready"
		end
		if t.state == "ready" then
			if t.delay == false then
				t.obj:queueLuaCommand("aitt.setMode('disabled')")
			elseif t.delay == true then
				if queue > 0 then
					setSpawn(t)
					queue = -2 -- artificial delay for next spawn test; prevents many multiple teleports and lag (is this needed anymore?)
				end
			end
			t.delay = true
		end
		t.dist.prev = t.dist.now
		if t.action ~= "chase" and t.action ~= "flee" then
			if t.state ~= "set" and trafficMode == "random" then
				if smartSpeed == true then
					t.speed = getSmartSpeed(t)
				else
					t.speed = speedLimit * t.road.score
				end
				if damageMode == true then
					if t.damage.now - t.damage.init >= damageUnit * 2 then -- traffic cars try to stop driving if heavily damaged
						t.speed = 0
					end
					if t.damage.now - t.damage.init >= damageUnit * 4 then
						if t.damage.bad == false and math.random(1, 10) == 10 then
							t.respawn = 0.8 -- delays respawn in case the player wants to look back at the site of the wreck
							t.lane = 1.5
							t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
							t.obj:queueLuaCommand("electrics.set_warn_signal(1)")
							t.damage.bad = true
						end
					end
				end
			else
				t.speed = 100
			end
			if t.speed <= crawlSpeed * 2 then -- waiting drivers
				t.timer.wait = t.timer.wait + timeTick
			else
				t.timer.wait = 0
			end
		else
			t.speed = 100
		end
		if t.lane == 0 and math.random(1, 2000) == 1 then
			t.lane = 1
			t.obj:queueLuaCommand("aitt.setLane("..t.lane..")")
			t.obj:queueLuaCommand("electrics.toggle_right_signal()")
		end
		if t.action ~= "none" then
			chasePlayer(t)
		end
		if (math.floor(t.timer.active * 10) / 10) % 1 == 0 then -- update road data every second
			updateRoad(t)
		end
		t.obj:queueLuaCommand("aitt.setSpeed("..tostring(t.speed)..")") -- main speed setter
		t.timer.active = t.timer.active + timeTick
	end
	if policeReady == true and policeAlert == true then
		simulatePolice()
	end
	queue = queue + 1
	if queue > #traffic then
		queue = 1
	end
end

local function onPreRender(dt, dtSim)
	loopTimer = loopTimer + dt
	if trafficReady == true then
		for i, t in ipairs(traffic) do
			t.obj:queueLuaCommand("aitt.updateGFX("..dtSim..")")
		end
	end
	if loopTimer >= timeTick then -- runs on ticks instead of real time; similar to onRaceTick in scenarios (future: remove this method)
		validateVehicles()
		if trafficReady == true then
			simulateTraffic()
		end
		loopTimer = 0
	end
end

M.onPreRender = onPreRender

M.ready = ready
M.setQuantity = setQuantity
M.setGroup = setGroup
M.spawnVehicles = spawnVehicles
M.deleteVehicles = deleteVehicles
M.activateVehicles = activateVehicles
M.setSpeedLimit = setSpeedLimit
M.setRespawnRate = setRespawnRate
M.setColorMode = setColorMode
M.setAiMode = setAiMode
M.setCustom = setCustom
M.setAsPolice = setAsPolice
M.toggleUnits = toggleUnits
M.togglePolice = togglePolice
M.toggleHeadlights = toggleHeadlights
M.toggleDamageKill = toggleDamageKill
M.toggleAwareness = toggleAwareness
M.toggleMethod = toggleMethod

return M